from functor_ninja import Try, Success, Fail


def test_success():
    result = Try(1).map(lambda v: v + 1)

    assert isinstance(result, Success)
    assert not isinstance(result, Fail)


def test_fail():
    result = Try(0).map(lambda v: 1 / v)

    assert isinstance(result, Fail)
    assert not isinstance(result, Success)


def test_of_success():
    result = Try.of(lambda: 0 / 1)

    assert isinstance(result, Success)
    assert not isinstance(result, Fail)


def test_of_fail():
    result = Try.of(lambda: 1 / 0)

    assert isinstance(result, Fail)
    assert not isinstance(result, Success)



from functor_ninja.monad import (
    A,
    B,
    Callable,
    Monad,
)

from typing import List as BaseList


class List(Monad[A]):
    def __init__(self, values: BaseList[A]):
        self.values = values

    @staticmethod
    def of(init: Callable[[], BaseList[A]]) -> "List[A]":
        values = init()
        return List(values)

    def len(self) -> int:
        return len(self.values)

    def map(self, f: Callable[[A], B]) -> "List[B]":
        result = [f(value) for value in self.values]
        return List(result)

    def flat_map(self, f: Callable[[A], "List[B]"]) -> "List[B]":
        result = [
            child
            for value in self.values
            for child in f(value).values
        ]
        return List(result)



from abc import abstractmethod
from typing import Callable, Generic, TypeVar

A = TypeVar("A")
B = TypeVar("B")


class Functor(Generic[A]):
    @abstractmethod
    def map(self, f: Callable[[A], B]) -> "Functor[B]":
        pass


class Monad(Functor[A]):
    @staticmethod
    @abstractmethod
    def of(init: Callable[[], A]) -> "Monad[A]":
        pass

    @abstractmethod
    def flat_map(self, f: Callable[[A], "Monad[B]"]) -> "Monad[B]":
        pass




from functor_ninja.monad import (
    A,
    B,
    Callable,
    Monad,
)


class Try(Monad[A]):
    def __init__(self, value: A):
        self.value = value

    @staticmethod
    def of(init: Callable[[], A]) -> "Try[A]":
        result = Try(None).map(lambda _: init())
        return result

    def map(self, f: Callable[[A], B]) -> "Try[B]":
        try:
            result = f(self.value)
            return Success(result)
        except Exception as e:
            return Fail(e)

    def flat_map(self, f: Callable[[A], "Try[B]"]) -> "Try[B]":
        try:
            result = f(self.value)
            return result
        except Exception as e:
            return Fail(e)


class Success(Try[A]):
    pass


class Fail(Try[Exception]):
    def map(self, f: Callable[[A], B]) -> "Fail[B]":
        return self

    def flat_map(self, f: Callable[[A], "Monad[B]"]) -> "Fail[B]":
        return self



from functor_ninja import List


def test_empty():
    empty = List([]).map(lambda v: v)

    assert empty.len() == 0


def test_multi():
    result = List([1, 2, 3]).map(lambda v: v + 1)

    assert result.values == [2, 3, 4]


def test_flatten():
    result = List([1, 2, 3]).flat_map(lambda v: List([v]))

    assert result.len() == 3

def test_flat_expend():
    result = List([1, 2, 3]).flat_map(lambda v: List([v]*v))

    assert result.len() == 6

def test_flat_compose():
    a = List(["a1", "a2", "a3"])
    b = List(["b1", "b2", "b3"])
    result = a.flat_map(
        lambda a:
        b.map(lambda b: (a, b))
    )

    assert result.len() == 9
